from bbot.modules.templates.postman import postman


class postman(postman):
    watched_events = ["ORG_STUB", "SOCIAL"]
    produced_events = ["CODE_REPOSITORY"]
    flags = ["passive", "subdomain-enum", "safe", "code-enum"]
    meta = {
        "description": "Query Postman's API for related workspaces, collections, requests and download them",
        "created_date": "2024-09-07",
        "author": "@domwhewell-sage",
    }

    reject_wildcards = False

    async def handle_event(self, event):
        # Handle postman profile
        if event.type == "SOCIAL":
            await self.handle_profile(event)
        elif event.type == "ORG_STUB":
            await self.handle_org_stub(event)

    async def handle_profile(self, event):
        profile_name = event.data.get("profile_name", "")
        self.verbose(f"Searching for postman workspaces, collections, requests belonging to {profile_name}")
        for item in await self.query(profile_name):
            workspace = item["document"]
            name = workspace["slug"]
            profile = workspace["publisherHandle"]
            if profile_name.lower() == profile.lower():
                self.verbose(f"Got {name}")
                workspace_url = f"{self.html_url}/{profile}/{name}"
                await self.emit_event(
                    {"url": workspace_url},
                    "CODE_REPOSITORY",
                    tags="postman",
                    parent=event,
                    context=f'{{module}} searched postman.com for workspaces belonging to "{profile_name}" and found "{name}" at {{event.type}}: {workspace_url}',
                )

    async def handle_org_stub(self, event):
        org_name = event.data
        self.verbose(f"Searching for any postman workspaces, collections, requests for {org_name}")
        for item in await self.query(org_name):
            workspace = item["document"]
            name = workspace["slug"]
            profile = workspace["publisherHandle"]
            self.verbose(f"Got {name}")
            workspace_url = f"{self.html_url}/{profile}/{name}"
            await self.emit_event(
                {"url": workspace_url},
                "CODE_REPOSITORY",
                tags="postman",
                parent=event,
                context=f'{{module}} searched postman.com for "{org_name}" and found matching workspace "{name}" at {{event.type}}: {workspace_url}',
            )

    async def query(self, query):
        data = []
        url = f"{self.base_url}/ws/proxy"
        json = {
            "service": "search",
            "method": "POST",
            "path": "/search-all",
            "body": {
                "queryIndices": [
                    "collaboration.workspace",
                ],
                "queryText": self.helpers.quote(query),
                "size": 100,
                "from": 0,
                "clientTraceId": "",
                "requestOrigin": "srp",
                "mergeEntities": "true",
                "nonNestedRequests": "true",
                "domain": "public",
            },
        }
        r = await self.helpers.request(url, method="POST", json=json, headers=self.headers)
        if r is None:
            return data
        status_code = getattr(r, "status_code", 0)
        try:
            json = r.json()
        except Exception as e:
            self.warning(f"Failed to decode JSON for {r.url} (HTTP status: {status_code}): {e}")
            return None
        return json.get("data", [])



from extractous import Extractor

from bbot.modules.base import BaseModule


class extractous(BaseModule):
    watched_events = ["FILESYSTEM"]
    produced_events = ["RAW_TEXT"]
    flags = ["passive", "safe"]
    meta = {
        "description": "Module to extract data from files",
        "created_date": "2024-06-03",
        "author": "@domwhewell-sage",
    }
    options = {
        "extensions": [
            "bak",  #  Backup File
            "bash",  #  Bash Script or Configuration
            "bashrc",  #  Bash Script or Configuration
            "conf",  #  Configuration File
            "cfg",  #  Configuration File
            "crt",  #  Certificate File
            "csv",  #  Comma Separated Values File
            "db",  #  SQLite Database File
            "sqlite",  #  SQLite Database File
            "doc",  #  Microsoft Word Document (Old Format)
            "docx",  #  Microsoft Word Document
            "ica",  #  Citrix Independent Computing Architecture File
            "indd",  #  Adobe InDesign Document
            "ini",  #  Initialization File
            "json",  #  JSON File
            "key",  #  Private Key File
            "pub",  #  Public Key File
            "log",  #  Log File
            "markdown",  #  Markdown File
            "md",  #  Markdown File
            "odg",  #  OpenDocument Graphics (LibreOffice, OpenOffice)
            "odp",  #  OpenDocument Presentation (LibreOffice, OpenOffice)
            "ods",  #  OpenDocument Spreadsheet (LibreOffice, OpenOffice)
            "odt",  #  OpenDocument Text (LibreOffice, OpenOffice)
            "pdf",  #  Adobe Portable Document Format
            "pem",  #  Privacy Enhanced Mail (SSL certificate)
            "pps",  #  Microsoft PowerPoint Slideshow (Old Format)
            "ppsx",  #  Microsoft PowerPoint Slideshow
            "ppt",  #  Microsoft PowerPoint Presentation (Old Format)
            "pptx",  #  Microsoft PowerPoint Presentation
            "ps1",  #  PowerShell Script
            "rdp",  #  Remote Desktop Protocol File
            "rsa",  #  RSA Private Key File
            "sh",  #  Shell Script
            "sql",  #  SQL Database Dump
            "swp",  #  Swap File (temporary file, often Vim)
            "sxw",  #  OpenOffice.org Writer document
            "txt",  #  Plain Text Document
            "vbs",  #  Visual Basic Script
            "wpd",  #  WordPerfect Document
            "xls",  #  Microsoft Excel Spreadsheet (Old Format)
            "xlsx",  #  Microsoft Excel Spreadsheet
            "xml",  #  eXtensible Markup Language File
            "yml",  #  YAML Ain't Markup Language
            "yaml",  #  YAML Ain't Markup Language
        ],
    }
    options_desc = {
        "extensions": "File extensions to parse",
    }

    deps_pip = ["extractous"]
    scope_distance_modifier = 1

    async def setup(self):
        self.extensions = list({e.lower().strip(".") for e in self.config.get("extensions", [])})
        return True

    async def filter_event(self, event):
        if "file" in event.tags:
            if not any(event.data["path"].endswith(f".{ext}") for ext in self.extensions):
                return False, "File extension not in the allowed list"
        else:
            return False, "Event is not a file"
        return True

    async def handle_event(self, event):
        file_path = event.data["path"]
        content = await self.scan.helpers.run_in_executor_mp(extract_text, file_path)
        if isinstance(content, tuple):
            error, traceback = content
            self.error(f"Error extracting text from {file_path}: {error}")
            self.trace(traceback)
            return

        if content:
            raw_text_event = self.make_event(
                content,
                "RAW_TEXT",
                context=f"Extracted text from {file_path}",
                parent=event,
            )
            await self.emit_event(raw_text_event)


def extract_text(file_path):
    """
    extract_text Extracts plaintext from a document path using extractous.

    :param file_path: The path of the file to extract text from.
    :return: ASCII-encoded plaintext extracted from the document.
    """

    try:
        extractor = Extractor()
        reader, metadata = extractor.extract_file(str(file_path))

        result = ""
        buffer = reader.read(4096)
        while len(buffer) > 0:
            result += buffer.decode("utf-8", errors="ignore")
            buffer = reader.read(4096)

        return result.strip()
    except Exception as e:
        import traceback

        return (str(e), traceback.format_exc())



from .base import ModuleTestBase


class TestAggregate(ModuleTestBase):
    config_overrides = {"dns": {"minimal": False}, "scope": {"report_distance": 1}}

    async def setup_before_prep(self, module_test):
        await module_test.mock_dns({"blacklanternsecurity.com": {"A": ["1.2.3.4"]}})

    def check(self, module_test, events):
        filename = next(module_test.scan.home.glob("scan-stats-table*.txt"))
        with open(filename) as f:
            assert "| A  " in f.read()



from .base import ModuleTestBase


class TestGithub_Codesearch(ModuleTestBase):
    config_overrides = {
        "modules": {
            "github_codesearch": {"api_key": "asdf", "limit": 1},
            "trufflehog": {"only_verified": False},
        },
        "omit_event_types": [],
        "scope": {"report_distance": 2},
    }
    modules_overrides = ["github_codesearch", "httpx", "trufflehog"]

    github_file_endpoint = (
        "/projectdiscovery/nuclei/06f242e5fce3439b7418877676810cbf57934875/v2/cmd/cve-annotate/main.go"
    )
    github_file_url = f"http://127.0.0.1:8888{github_file_endpoint}"
    github_file_content = """-----BEGIN PRIVATE KEY-----
MIICdwIBADANBgkqhkiG9w0BAQEFAASCAmEwggJdAgEAAoGBAOBY2pd9PSQvuxqu
WXFNVgILTWuUc721Wc2sFNvp4beowhUe1lfxaq5ZfCJcz7z4QsqFhOeks69O9UIb
oiOTDocPDog9PHO8yZXopHm0StFZvSjjKSNuFvy/WopPTGpxUZ5boCaF1CXumY7W
FL+jIap5faimLL9prIwaQKBwv80lAgMBAAECgYEAxvpHtgCgD849tqZYMgOTevCn
U/kwxltoMOClB39icNA+gxj8prc6FTTMwnVq0oGmS5UskX8k1yHCqUV1AvRU9o+q
I8L8a3F3TQKQieI/YjiUNK8A87bKkaiN65ooOnhT+I3ZjZMPR5YEyycimMp22jsv
LyX/35J/wf1rNiBs/YECQQDvtxgmMhE+PeajXqw1w2C3Jds27hI3RPDnamEyWr/L
KkSplbKTF6FuFDYOFdJNPrfxm1tx2MZ2cBfs+h/GnCJVAkEA75Z9w7q8obbqGBHW
9bpuFvLjW7bbqO7HBuXYX9zQcZL6GSArFP0ba5lhgH1qsVQfxVWVyiV9/chme7xc
ljfvkQJBAJ7MpSPQcRnRefNp6R0ok+5gFqt55PlWI1y6XS81bO7Szm+laooE0n0Q
yIpmLE3dqY9VgquVlkupkD/9poU0s40CQD118ZVAVht1/N9n1Cj9RjiE3mYspnTT
rCLM25Db6Gz6M0Y2xlaAB4S2uBhqE/Chj/TjW6WbsJJl0kRzsZynhMECQFYKiM1C
T4LB26ynW00VE8z4tEWSoYt4/Vn/5wFhalVjzoSJ8Hm2qZiObRYLQ1m0X4KnkShk
Gnl54dJHT+EhlfY=
-----END PRIVATE KEY-----"""

    async def setup_before_prep(self, module_test):
        expect_args = {"method": "GET", "uri": self.github_file_endpoint}
        respond_args = {"response_data": self.github_file_content}
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

        module_test.httpx_mock.add_response(url="https://api.github.com/zen")
        module_test.httpx_mock.add_response(
            url="https://api.github.com/search/code?per_page=100&type=Code&q=blacklanternsecurity.com&page=1",
            json={
                "total_count": 214,
                "incomplete_results": False,
                "items": [
                    {
                        "html_url": "https://github.com/projectdiscovery/nuclei/blob/06f242e5fce3439b7418877676810cbf57934875/v2/cmd/cve-annotate/main.go",
                        "repository": {
                            "html_url": "https://github.com/projectdiscovery/nuclei",
                        },
                    },
                    {
                        "html_url": "https://github.com/projectdiscovery/nuclei/blob/06f242e5fce3439b7418877676810cbf57934875/v2/cmd/cve-annotate/main.go2",
                        "repository": {
                            "html_url": "https://github.com/projectdiscovery/nuclei",
                        },
                    },
                    {
                        "html_url": "https://github.com/projectdiscovery/nuclei/blob/06f242e5fce3439b7418877676810cbf57934875/v2/cmd/cve-annotate/main.go3",
                        "repository": {
                            "html_url": "https://github.com/projectdiscovery/nuclei",
                        },
                    },
                ],
            },
        )

    async def setup_after_prep(self, module_test):
        module_test.module.github_raw_url = "http://127.0.0.1:8888/"

    def check(self, module_test, events):
        assert 1 == len([e for e in events if e.type == "URL_UNVERIFIED"])
        assert 1 == len(
            [
                e
                for e in events
                if e.type == "URL_UNVERIFIED" and e.data == self.github_file_url and e.scope_distance == 2
            ]
        ), "Failed to emit URL_UNVERIFIED"
        assert 1 == len(
            [
                e
                for e in events
                if e.type == "CODE_REPOSITORY"
                and "git" in e.tags
                and e.data["url"] == "https://github.com/projectdiscovery/nuclei"
                and e.scope_distance == 1
            ]
        ), "Failed to emit CODE_REPOSITORY"
        assert 1 == len(
            [e for e in events if e.type == "URL" and e.data == self.github_file_url and e.scope_distance == 2]
        ), "Failed to visit URL"
        assert 1 == len(
            [
                e
                for e in events
                if e.type == "HTTP_RESPONSE" and e.data["url"] == self.github_file_url and e.scope_distance == 2
            ]
        ), "Failed to visit URL"
        assert [
            e for e in events if e.type == "FINDING" and str(e.module) == "trufflehog"
        ], "Failed to find secret in repo file"



from .base import ModuleTestBase


class TestNucleiManual(ModuleTestBase):
    targets = ["http://127.0.0.1:8888"]
    modules_overrides = ["httpx", "excavate", "nuclei"]
    config_overrides = {
        "web": {
            "spider_distance": 1,
            "spider_depth": 1,
        },
        "modules": {
            "nuclei": {
                "version": "2.9.4",
                "mode": "manual",
                "concurrency": 2,
                "ratelimit": 10,
                "templates": "/tmp/.bbot_test/tools/nuclei-templates/http/miscellaneous/",
                "interactsh_disable": True,
                "directory_only": False,
            }
        },
    }

    test_html = """
    html>
 <head>
  <title>Index of /test</title>
 </head>
 <body>
<h1>Index of /test</h1>
  <table>
   <tr><th><a href="?C=N;O=D">Name</a></th><th><a href="?C=M;O=A">Last modified</a></th><th><a href="?C=S;O=A">Size</a></th></tr>
   <tr><th colspan="3"><hr></th></tr>
<tr><td><a href="/">Parent Directory</a></td><td>&nbsp;</td><td align="right">  - </td></tr>
</table>
<address>Apache/2.4.38 (Debian) Server at http://127.0.0.1:8888/testmultipleruns.html</address>
</body></html>
"""

    async def setup_after_prep(self, module_test):
        expect_args = {"method": "GET", "uri": "/"}
        respond_args = {"response_data": self.test_html}
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)
        expect_args = {"method": "GET", "uri": "/testmultipleruns.html"}
        respond_args = {"response_data": "<html>Copyright 1984</html>"}
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

    def check(self, module_test, events):
        first_run_detect = False
        second_run_detect = False
        for e in events:
            if e.type == "FINDING":
                if "Directory listing enabled" in e.data["description"]:
                    first_run_detect = True
                elif "Copyright" in e.data["description"]:
                    second_run_detect = True
        assert first_run_detect
        assert second_run_detect


class TestNucleiSevere(TestNucleiManual):
    modules_overrides = ["httpx", "nuclei"]
    config_overrides = {
        "modules": {
            "nuclei": {
                "mode": "severe",
                "concurrency": 1,
                "templates": "/tmp/.bbot_test/tools/nuclei-templates/vulnerabilities/generic/generic-linux-lfi.yaml",
            }
        },
        "interactsh_disable": True,
    }

    async def setup_after_prep(self, module_test):
        expect_args = {"method": "GET", "uri": "/etc/passwd"}
        respond_args = {"response_data": "<html>root:.*:0:0:</html>"}
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

    def check(self, module_test, events):
        assert any(
            e.type == "VULNERABILITY" and "Generic Linux - Local File Inclusion" in e.data["description"]
            for e in events
        )


class TestNucleiTechnology(TestNucleiManual):
    config_overrides = {
        "interactsh_disable": True,
        "modules": {"nuclei": {"mode": "technology", "concurrency": 2, "tags": "apache"}},
    }

    async def setup_before_prep(self, module_test):
        expect_args = {"method": "GET", "uri": "/"}
        respond_args = {
            "response_data": "<html><Directory></Directory></html>",
            "headers": {"Server": "Apache/2.4.52 (Ubuntu)"},
        }
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

    def check(self, module_test, events):
        assert any(e.type == "TECHNOLOGY" and "apache" in e.data["technology"].lower() for e in events)

        with open(module_test.scan.home / "debug.log") as f:
            assert "Using Interactsh Server" not in f.read()


class TestNucleiBudget(TestNucleiManual):
    config_overrides = {
        "modules": {
            "nuclei": {
                "mode": "budget",
                "concurrency": 1,
                "tags": "spiderfoot",
                "templates": "/tmp/.bbot_test/tools/nuclei-templates/exposed-panels/spiderfoot.yaml",
                "interactsh_disable": True,
            }
        }
    }

    async def setup_before_prep(self, module_test):
        expect_args = {"method": "GET", "uri": "/"}
        respond_args = {"response_data": "<html><title>SpiderFoot</title><p>support@spiderfoot.net</p></html>"}
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

    def check(self, module_test, events):
        assert any(e.type == "TECHNOLOGY" and "spider" in e.data["technology"] for e in events)


class TestNucleiRetries(TestNucleiManual):
    config_overrides = {
        "interactsh_disable": True,
        "modules": {"nuclei": {"tags": "musictraveler"}},
    }

    async def setup_before_prep(self, module_test):
        expect_args = {"method": "GET", "uri": "/"}
        respond_args = {
            "response_data": "content",
        }
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

    def check(self, module_test, events):
        with open(module_test.scan.home / "debug.log") as f:
            assert "-retries 0" in f.read()


class TestNucleiRetriesCustom(TestNucleiRetries):
    config_overrides = {
        "interactsh_disable": True,
        "modules": {"nuclei": {"tags": "musictraveler", "retries": 1}},
    }

    def check(self, module_test, events):
        with open(module_test.scan.home / "debug.log") as f:
            assert "-retries 1" in f.read()


class TestNucleiCustomHeaders(TestNucleiManual):
    custom_headers = {"testheader1": "test1", "testheader2": "test2"}
    config_overrides = TestNucleiManual.config_overrides
    config_overrides["web"]["http_headers"] = custom_headers

    async def setup_after_prep(self, module_test):
        expect_args = {"method": "GET", "uri": "/", "headers": self.custom_headers}
        respond_args = {"response_data": self.test_html}
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)
        expect_args = {"method": "GET", "uri": "/testmultipleruns.html", "headers": {"nonexistent": "nope"}}
        respond_args = {"response_data": "<html>Copyright 1984</html>"}
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

    def check(self, module_test, events):
        first_run_detect = False
        second_run_detect = False
        for e in events:
            if e.type == "FINDING":
                if "Directory listing enabled" in e.data["description"]:
                    first_run_detect = True
                elif "Copyright" in e.data["description"]:
                    second_run_detect = True
        # we should find the first one because it requires our custom headers
        assert first_run_detect
        # the second one requires different headers, so we shouldn't find it
        assert not second_run_detect



import time

from bbot.modules.base import BaseModule


class dnsbrute_mutations(BaseModule):
    flags = ["subdomain-enum", "active", "aggressive", "slow"]
    watched_events = ["DNS_NAME"]
    produced_events = ["DNS_NAME"]
    meta = {
        "description": "Brute-force subdomains with massdns + target-specific mutations",
        "author": "@TheTechromancer",
        "created_date": "2024-04-25",
    }
    options = {
        "max_mutations": 100,
    }
    options_desc = {
        "max_mutations": "Maximum number of target-specific mutations to try per subdomain",
    }
    deps_common = ["massdns"]
    _qsize = 10000

    async def setup(self):
        self.found = {}
        self.parent_events = {}
        self.max_mutations = self.config.get("max_mutations", 500)
        # 800M bits == 100MB bloom filter == 10M entries before false positives start emerging
        self.mutations_tried = self.helpers.bloom_filter(800000000)
        self._mutation_run_counter = {}
        return True

    async def handle_event(self, event):
        # here we don't brute-force, we just add the subdomain to our end-of-scan
        host = str(event.host)
        self.parent_events[host] = event
        if self.helpers.is_subdomain(host):
            subdomain, domain = host.split(".", 1)
            if not self.helpers.dns.brute.has_excessive_digits(subdomain):
                try:
                    self.found[domain].add(subdomain)
                except KeyError:
                    self.found[domain] = {subdomain}

    async def get_parent_event(self, subdomain):
        start = time.time()
        parent_host = await self.helpers.run_in_executor(self.helpers.closest_match, subdomain, self.parent_events)
        elapsed = time.time() - start
        self.trace(f"{subdomain}: got closest match among {len(self.parent_events):,} parent events in {elapsed:.2f}s")
        return self.parent_events[parent_host]

    async def finish(self):
        """
        TODO: speed up this loop.
            We should see if we can combine multiple runs together instead of running them each individually.
        """
        found = sorted(self.found.items(), key=lambda x: len(x[-1]), reverse=True)
        # if we have a lot of rounds to make, don't try mutations on less-populated domains
        trimmed_found = []
        if found:
            avg_subdomains = sum([len(subdomains) for domain, subdomains in found[:50]]) / len(found[:50])
            for i, (domain, subdomains) in enumerate(found):
                # accept domains that are in the top 50 or have more than 5 percent of the average number of subdomains
                if i < 50 or (len(subdomains) > 1 and len(subdomains) >= (avg_subdomains * 0.05)):
                    trimmed_found.append((domain, subdomains))
                else:
                    self.verbose(
                        f"Skipping mutations on {domain} because it only has {len(subdomains):,} subdomain(s) (avg: {avg_subdomains:,})"
                    )

        base_mutations = set()
        try:
            for i, (domain, subdomains) in enumerate(trimmed_found):
                self.verbose(f"{domain} has {len(subdomains):,} subdomains")
                # keep looping as long as we're finding things
                while 1:
                    query = domain

                    mutations = set(base_mutations)

                    def add_mutation(m):
                        h = f"{m}.{domain}"
                        if h not in self.mutations_tried:
                            self.mutations_tried.add(h)
                            mutations.add(m)

                    # try every subdomain everywhere else
                    for _domain, _subdomains in found:
                        if _domain == domain:
                            continue
                        for s in _subdomains:
                            first_segment = s.split(".")[0]
                            # skip stuff with lots of numbers (e.g. PTRs)
                            if self.helpers.dns.brute.has_excessive_digits(first_segment):
                                continue
                            add_mutation(first_segment)
                            for word in self.helpers.extract_words(
                                first_segment, word_regexes=self.helpers.word_cloud.dns_mutator.extract_word_regexes
                            ):
                                add_mutation(word)

                    # numbers + devops mutations
                    for mutation in self.helpers.word_cloud.mutations(
                        subdomains, cloud=False, numbers=3, number_padding=1
                    ):
                        for delimiter in ("", ".", "-"):
                            m = delimiter.join(mutation).lower()
                            add_mutation(m)

                    # special dns mutator
                    for subdomain in self.helpers.word_cloud.dns_mutator.mutations(
                        subdomains, max_mutations=self.max_mutations
                    ):
                        add_mutation(subdomain)

                    # skip if there's hardly any mutations
                    if len(mutations) < 10:
                        self.verbose(
                            f"Skipping {len(mutations):,} mutations against {domain} because there are less than 10"
                        )
                        break

                    if mutations:
                        self.info(f"Trying {len(mutations):,} mutations against {domain} ({i+1}/{len(trimmed_found)})")
                        results = await self.helpers.dns.brute(self, query, mutations)
                        try:
                            mutation_run = self._mutation_run_counter[domain]
                        except KeyError:
                            self._mutation_run_counter[domain] = mutation_run = 1
                        self._mutation_run_counter[domain] += 1
                        for hostname in results:
                            parent_event = await self.get_parent_event(hostname)
                            mutation_run_ordinal = self.helpers.integer_to_ordinal(mutation_run)
                            await self.emit_event(
                                hostname,
                                "DNS_NAME",
                                parent=parent_event,
                                tags=[f"mutation-{mutation_run}"],
                                abort_if=self.abort_if,
                                context=f'{{module}} found a mutated subdomain of "{parent_event.host}" on its {mutation_run_ordinal} run: {{event.type}}: {{event.data}}',
                            )
                        if results:
                            continue
                    break
        except AssertionError as e:
            self.warning(e)

    def abort_if(self, event):
        if not event.scope_distance == 0:
            return True, "event is not in scope"
        if "wildcard" in event.tags:
            return True, "event is a wildcard"
        if "unresolved" in event.tags:
            return True, "event is unresolved"
        return False, ""



import json
import asyncio
import websockets

from bbot.modules.output.base import BaseOutputModule


class Websocket(BaseOutputModule):
    watched_events = ["*"]
    meta = {"description": "Output to websockets", "created_date": "2022-04-15", "author": "@TheTechromancer"}
    options = {"url": "", "token": "", "preserve_graph": True}
    options_desc = {
        "url": "Web URL",
        "token": "Authorization Bearer token",
        "preserve_graph": "Preserve full chains of events in the graph (prevents orphans)",
    }

    async def setup(self):
        self.url = self.config.get("url", "")
        if not self.url:
            return False, "Must set URL"
        self.token = self.config.get("token", "")
        self._ws = None
        return True

    async def handle_event(self, event):
        event_json = event.json()
        await self.send(event_json)

    async def ws(self, rebuild=False):
        if self._ws is None or rebuild:
            kwargs = {"close_timeout": 0.5}
            if self.token:
                kwargs.update({"extra_headers": {"Authorization": f"Bearer {self.token}"}})
            verbs = ("Building", "Built")
            if rebuild:
                verbs = ("Rebuilding", "Rebuilt")
            self.debug(f"{verbs[0]} websocket connection to {self.url}")
            self._ws = await websockets.connect(self.url, **kwargs)
            self.debug(f"{verbs[1]} websocket connection to {self.url}")
        return self._ws

    async def send(self, message):
        rebuild = False
        while not self.scan.stopped:
            try:
                ws = await self.ws(rebuild=rebuild)
                message_str = json.dumps(message)
                self.debug(f"Sending message of length {len(message_str)}")
                await ws.send(message_str)
                rebuild = False
                break
            except Exception as e:
                self.warning(f"Error sending message: {e}, retrying")
                await asyncio.sleep(1)
                rebuild = True

    async def cleanup(self):
        if self._ws is not None:
            self.debug(f"Closing connection to {self.url}")
            await self._ws.close()
            self.debug(f"Closed connection to {self.url}")
        self._ws = None



import asyncio
from contextlib import suppress

from bbot.modules.base import BaseInterceptModule


class ScanIngress(BaseInterceptModule):
    """
    This is always the first intercept module in the chain, responsible for basic scope checks

    It has its own incoming queue, but will also pull events from modules' outgoing queues
    """

    watched_events = ["*"]
    # accept all events regardless of scope distance
    scope_distance_modifier = None
    _name = "_scan_ingress"
    _qsize = -1

    @property
    def priority(self):
        # we are the highest priority
        return -99

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._module_priority_weights = None
        self._non_intercept_modules = None
        # track incoming duplicates module-by-module (for `suppress_dupes` attribute of modules)
        self.incoming_dup_tracker = set()

    async def init_events(self, events=None):
        """
        Initializes events by seeding the scanner with target events and distributing them for further processing.

        Notes:
            - This method populates the event queue with initial target events.
            - It also marks the Scan object as finished with initialization by setting `_finished_init` to True.
        """
        if events is None:
            events = self.scan.target.seeds.events
        async with self.scan._acatch(self.init_events), self._task_counter.count(self.init_events):
            sorted_events = sorted(events, key=lambda e: len(e.data))
            for event in [self.scan.root_event] + sorted_events:
                event._dummy = False
                event.web_spider_distance = 0
                event.scan = self.scan
                if event.parent is None:
                    event.parent = self.scan.root_event
                if event.module is None:
                    event.module = self.scan._make_dummy_module(name="TARGET", _type="TARGET")
                if event != self.scan.root_event:
                    event.discovery_context = f"Scan {self.scan.name} seeded with " + "{event.type}: {event.data}"
                self.verbose(f"Target: {event}")
                await self.queue_event(event, {})
            await asyncio.sleep(0.1)
            self.scan._finished_init = True

    async def handle_event(self, event, **kwargs):
        # don't accept dummy events
        if event._dummy:
            return False, "cannot emit dummy event"

        # don't accept events with self as parent
        if not event.type == "SCAN":
            if event == event.get_parent():
                return False, "event's parent is itself"
            if not event.discovery_context:
                self.warning(f"Event {event} has no discovery context")

        # don't accept duplicates
        if self.is_incoming_duplicate(event, add=True):
            if not event._graph_important:
                return False, "event was already emitted by its module"
            else:
                self.debug(
                    f"Event {event} was already emitted by its module, but it's graph-important so it gets a pass"
                )

        # update event's scope distance based on its parent
        event.scope_distance = event.parent.scope_distance + 1

        # special handling of URL extensions
        url_extension = getattr(event, "url_extension", None)
        if url_extension is not None:
            if url_extension in self.scan.url_extension_httpx_only:
                event.add_tag("httpx-only")
                event._omit = True

            # blacklist by extension
            if url_extension in self.scan.url_extension_blacklist:
                self.debug(
                    f"Blacklisting {event} because its extension (.{url_extension}) is blacklisted in the config"
                )
                event.add_tag("blacklisted")

        # main scan blacklist
        event_blacklisted = self.scan.blacklisted(event)

        # reject all blacklisted events
        if event_blacklisted or "blacklisted" in event.tags:
            return False, "event is blacklisted"

        # Scope shepherding
        # here is where we make sure in-scope events are set to their proper scope distance
        if event.host:
            event_whitelisted = self.scan.whitelisted(event)
            if event_whitelisted:
                self.debug(f"Making {event} in-scope because its main host matches the scan target")
                event.scope_distance = 0

        # nerf event's priority if it's not in scope
        event.module_priority += event.scope_distance

    @property
    def non_intercept_modules(self):
        if self._non_intercept_modules is None:
            self._non_intercept_modules = [m for m in self.scan.modules.values() if not m._intercept]
        return self._non_intercept_modules

    @property
    def incoming_queues(self):
        queues = [self.incoming_event_queue] + [m.outgoing_event_queue for m in self.non_intercept_modules]
        return [q for q in queues if q is not False]

    @property
    def module_priority_weights(self):
        if not self._module_priority_weights:
            # we subtract from six because lower priorities == higher weights
            priorities = [5] + [6 - m.priority for m in self.non_intercept_modules]
            self._module_priority_weights = priorities
        return self._module_priority_weights

    async def get_incoming_event(self):
        for q in self.helpers.weighted_shuffle(self.incoming_queues, self.module_priority_weights):
            try:
                return q.get_nowait()
            except (asyncio.queues.QueueEmpty, AttributeError):
                continue
        raise asyncio.queues.QueueEmpty()

    def is_incoming_duplicate(self, event, add=False):
        """
        Calculate whether an event is a duplicate in the context of the module that emitted it
        This will return True if the event's parent module has raised the event before.
        """
        try:
            event_hash = event.module._outgoing_dedup_hash(event)
        except AttributeError:
            module_name = str(getattr(event, "module", ""))
            event_hash = hash((event, module_name))
        is_dup = event_hash in self.incoming_dup_tracker
        if add:
            self.incoming_dup_tracker.add(event_hash)
        suppress_dupes = getattr(event.module, "suppress_dupes", True)
        if suppress_dupes and is_dup:
            return True
        return False


class ScanEgress(BaseInterceptModule):
    """
    This is always the last intercept module in the chain, responsible for executing and acting on the
    `abort_if` and `on_success_callback` functions.
    """

    watched_events = ["*"]
    # accept all events regardless of scope distance
    scope_distance_modifier = None
    _name = "_scan_egress"

    @property
    def priority(self):
        # we are the lowest priority
        return 99

    async def handle_event(self, event, **kwargs):
        abort_if = kwargs.pop("abort_if", None)
        on_success_callback = kwargs.pop("on_success_callback", None)

        # omit certain event types
        if event.type in self.scan.omitted_event_types:
            if "target" in event.tags:
                self.debug(f"Allowing omitted event: {event} because it's a target")
            else:
                event._omit = True

        # make event internal if it's above our configured report distance
        event_in_report_distance = event.scope_distance <= self.scan.scope_report_distance
        event_will_be_output = event.always_emit or event_in_report_distance

        if not event_will_be_output:
            self.debug(
                f"Making {event} internal because its scope_distance ({event.scope_distance}) > scope_report_distance ({self.scan.scope_report_distance})"
            )
            event.internal = True

        if event.type in self.scan.omitted_event_types:
            self.debug(f"Omitting {event} because its type is omitted in the config")
            event._omit = True

        # if we discovered something interesting from an internal event,
        # make sure we preserve its chain of parents
        parent = event.parent
        event_is_graph_worthy = (not event.internal) or event._graph_important
        parent_is_graph_worthy = (not parent.internal) or parent._graph_important
        if event_is_graph_worthy and not parent_is_graph_worthy:
            parent_in_report_distance = parent.scope_distance <= self.scan.scope_report_distance
            if parent_in_report_distance:
                parent.internal = False
            if not parent._graph_important:
                parent._graph_important = True
                self.debug(f"Re-queuing internal event {parent} with parent {event} to prevent graph orphan")
                await self.emit_event(parent)

        if event._suppress_chain_dupes:
            for parent in event.get_parents():
                if parent == event:
                    return False, f"an identical parent {event} was found, and _suppress_chain_dupes=True"

        # custom callback - abort event emission it returns true
        abort_result = False
        if callable(abort_if):
            async with self.scan._acatch(context=abort_if):
                abort_result = await self.scan.helpers.execute_sync_or_async(abort_if, event)
            msg = f"{event.module}: not raising event {event} due to custom criteria in abort_if()"
            with suppress(ValueError, TypeError):
                abort_result, reason = abort_result
                msg += f": {reason}"
            if abort_result:
                return False, msg

        # run success callback before distributing event (so it can add tags, etc.)
        if callable(on_success_callback):
            async with self.scan._acatch(context=on_success_callback):
                await self.scan.helpers.execute_sync_or_async(on_success_callback, event)

    async def forward_event(self, event, kwargs):
        """
        Queue event with modules
        """
        # absorb event into the word cloud if it's in scope
        if -1 < event.scope_distance < 1:
            self.scan.word_cloud.absorb_event(event)

        for mod in self.scan.modules.values():
            # don't distribute events to intercept modules
            if not mod._intercept:
                await mod.queue_event(event)



import time
import asyncio

from bbot.modules.templates.subdomain_enum import subdomain_enum_apikey


class SubdomainRadar(subdomain_enum_apikey):
    watched_events = ["DNS_NAME"]
    produced_events = ["DNS_NAME"]
    flags = ["subdomain-enum", "passive", "safe"]
    meta = {
        "description": "Query the Subdomain API for subdomains",
        "created_date": "2022-07-08",
        "author": "@TheTechromancer",
        "auth_required": True,
    }
    options = {"api_key": "", "group": "fast", "timeout": 120}
    options_desc = {
        "api_key": "SubDomainRadar.io API key",
        "group": "The enumeration group to use. Choose from fast, medium, deep",
        "timeout": "Timeout in seconds",
    }

    base_url = "https://api.subdomainradar.io"
    ping_url = f"{base_url}/profile"
    group_choices = ("fast", "medium", "deep")

    # set this really high so the poll loop finishes as soon as possible
    _qsize = 9999999

    async def setup(self):
        self.group = self.config.get("group", "fast").strip().lower()
        self.timeout = self.config.get("timeout", 120)
        if self.group not in self.group_choices:
            return False, f'Invalid group: "{self.group}", please choose from {",".join(self.group_choices)}'
        success, reason = await self.require_api_key()
        if not success:
            return success, reason
        # convert groups to enumerators
        enumerators = {}
        response = await self.api_request(f"{self.base_url}/enumerators/groups")
        status_code = getattr(response, "status_code", 0)
        if status_code != 200:
            return False, f"Failed to get enumerators: (HTTP status code: {status_code})"
        else:
            try:
                j = response.json()
            except Exception:
                return False, "Failed to get enumerators: failed to parse response as JSON"
            for group in j:
                group_name = group.get("name", "").strip().lower()
                if group_name:
                    group_enumerators = []
                    for enumerator in group.get("enumerators", []):
                        enumerator_name = enumerator.get("display_name", "")
                        if enumerator_name:
                            group_enumerators.append(enumerator_name)
                    if group_enumerators:
                        enumerators[group_name] = group_enumerators

        self.enumerators = enumerators.get(self.group, [])
        if not self.enumerators:
            return False, f'No enumerators found for group: "{self.group}" ({self.enumerators})'

        self.enum_tasks = {}
        self.poll_task = asyncio.create_task(self.task_poll_loop())

        return True

    def prepare_api_request(self, url, kwargs):
        if self.api_key:
            kwargs["headers"] = {"Authorization": f"Bearer {self.api_key}"}
        return url, kwargs

    async def handle_event(self, event):
        query = self.make_query(event)
        # start enumeration task
        url = f"{self.base_url}/enumerate"
        response = await self.api_request(
            url, method="POST", json={"domains": [query], "enumerators": self.enumerators}
        )
        try:
            j = response.json()
        except Exception:
            self.warning(f"Failed to parse response as JSON: {getattr(response, 'text', '')}")
            return
        task_id = j.get("tasks", {}).get(query, "")
        if not task_id:
            self.warning(f"Failed to start enumeration for {query}")
            return
        self.enum_tasks[query] = (task_id, time.time(), event)
        self.debug(f"Started enumeration task for {query}; task id: {task_id}")

    async def task_poll_loop(self):
        # async with self._task_counter.count(f"{self.name}.task_poll_loop()"):
        while 1:
            for query, (task_id, start_time, event) in list(self.enum_tasks.items()):
                url = f"{self.base_url}/tasks/{task_id}"
                response = await self.api_request(url)
                if getattr(response, "status_code", 0) == 200:
                    finished = await self.parse_response(response, query, event)
                    if finished:
                        self.enum_tasks.pop(query)
                        continue
                # if scan is finishing, consider timeout
                if self.scan.status == "FINISHING":
                    if start_time + self.timeout < time.time():
                        self.enum_tasks.pop(query)
                        self.info(f"Enumeration task for {query} timed out")

            if self.scan.status == "FINISHING" and not self.enum_tasks:
                break
            await self.helpers.sleep(5)

    async def parse_response(self, response, query, event):
        j = response.json()
        status = j.get("status", "")
        if status.lower() == "completed":
            for subdomain in j.get("subdomains", []):
                hostname = subdomain.get("subdomain", "")
                if hostname and hostname.endswith(f".{query}") and not hostname == event.data:
                    await self.emit_event(
                        hostname,
                        "DNS_NAME",
                        event,
                        abort_if=self.abort_if,
                        context=f'{{module}} searched SubDomainRadar.io API for "{query}" and found {{event.type}}: {{event.data}}',
                    )
            return True
        return False

    async def finish(self):
        start_time = time.time()
        while self.enum_tasks and not self.poll_task.done():
            elapsed_time = time.time() - start_time
            if elapsed_time >= self.timeout:
                self.warning(f"Timed out waiting for the following tasks to finish: {self.enum_tasks}")
                for query, (task_id, _, _) in list(self.enum_tasks.items()):
                    url = f"{self.base_url}/tasks/{task_id}"
                    self.warning(f"    - {query} ({url})")
                break

            self.verbose(
                f"Waiting for enumeration task poll loop to finish ({int(elapsed_time)}/{self.timeout} seconds)"
            )

            try:
                # Wait for the task to complete or for 10 seconds, whichever comes first
                await asyncio.wait_for(asyncio.shield(self.poll_task), timeout=10)
            except asyncio.TimeoutError:
                # This just means our 10-second check has elapsed, not that the task failed
                pass

        # Cancel the poll_task if it's still running
        if not self.poll_task.done():
            self.poll_task.cancel()
            try:
                await self.poll_task
            except asyncio.CancelledError:
                pass



import sqlite3
from .base import ModuleTestBase


class TestSQLite(ModuleTestBase):
    targets = ["evilcorp.com"]

    def check(self, module_test, events):
        sqlite_output_file = module_test.scan.home / "output.sqlite"
        assert sqlite_output_file.exists(), "SQLite output file not found"
        with sqlite3.connect(sqlite_output_file) as db:
            cursor = db.cursor()
            results = cursor.execute("SELECT * FROM event").fetchall()
            assert len(results) == 3, "No events found in SQLite database"
            results = cursor.execute("SELECT * FROM scan").fetchall()
            assert len(results) == 1, "No scans found in SQLite database"
            results = cursor.execute("SELECT * FROM target").fetchall()
            assert len(results) == 1, "No targets found in SQLite database"



from bbot.modules.templates.bucket import bucket_template


class bucket_amazon(bucket_template):
    watched_events = ["DNS_NAME", "STORAGE_BUCKET"]
    produced_events = ["STORAGE_BUCKET", "FINDING"]
    flags = ["active", "safe", "cloud-enum", "web-basic"]
    meta = {
        "description": "Check for S3 buckets related to target",
        "created_date": "2022-11-04",
        "author": "@TheTechromancer",
    }
    options = {"permutations": False}
    options_desc = {
        "permutations": "Whether to try permutations",
    }
    scope_distance_modifier = 3

    cloud_helper_name = "amazon"
    delimiters = ("", ".", "-")
    base_domains = ["s3.amazonaws.com"]
    regions = [None]
    supports_open_check = True
